# coding: utf-8
"""
Django-rest-easy provides base classes for API views and serializers.

To leverage the QOL features of django-rest-easy, you should use the followint base classes for your serializers:
* :class:`rest_easy.serializers.Serializer`
* :class:`rest_easy.serializers.ModelSerializer`

And if it's model-based, it should use one of the base views provided in the :mod:`rest_easy.views`
module - preferably :class:`rest_easy.views.ModelViewSet` or :class:`rest_easy.views.ReadOnlyModelViewSet`.

As a baseline, all responses using django-rest-easy extension will contain top-level model and schema fields.

Guidelines regarding schemas are as usual: they have to be 100% backwards compatible. In the case of breaking changes,
a serializer with new schema should be created, and the old one slowly faded away - and removed only when no
applications use it - or when it's decided that the feature can't be supported anymore.

An alternative to multi-version fadeout is single-version fadeout, where the change is implemented as a set of
acceptable changes (that is, you can remove the old field only when all clients stop using it - even if it means
sending duplicate data for quite some time).

The classes from this module don't disable any behaviour inherent to Django Rest Framework - anything that is possible
there will be possible with the django-rest-easy base classes.

Uses followint settings:
REST_EASY_AUTOIMPORT_SERIALIZERS_FROM - for autoimporting serializers.
REST_EASY_VIEW_BASES - for prepending bases to generic views.
"""
from django.apps import AppConfig
from django.conf import settings

default_app_config = 'rest_easy.ApiConfig'


class ApiConfig(AppConfig):
    """
    AppConfig autoimporting serializers.

    It scans all installed applications for modules specified in settings.REST_EASY_AUTOIMPORT_SERIALIZERS_FROM
    parameter, trying to import them so that all residing serializers using
    :class:`rest_easy.serializers.SerializerCreator` metaclass will be added to
    :class:`rest_easy.registers.SerializerRegister`.

    In the case of a module not being present in app's context, the import is skipped.
    In the case of a module existing but failing to import, an exception will be raised.
    """
    name = 'rest_easy'
    label = 'rest_easy'

    default_paths = ['serializers', 'api.serializers']

    @property
    def paths(self):
        """
        Get import paths - from settings or defaults.
        """
        return getattr(settings, 'REST_EASY_AUTOIMPORT_SERIALIZERS_FROM', self.default_paths)

    def autodiscover(self):
        """
        Auto-discover serializers in installed apps, fail silently when not present, re-raise exception when present
        and import fails. Borrowed form django.contrib.admin with added nested presence check.
        """

        from importlib import import_module
        from django.utils.module_loading import module_has_submodule

        for app in settings.INSTALLED_APPS:
            mod = import_module(app)

            # Attempt to import the app's serializers.
            for item in self.paths:
                try:
                    import_module('{}.{}'.format(app, item))

                except (TypeError, ImportError):
                    # Decide whether to bubble up this error. If the app just
                    # doesn't have serializers module, we can ignore the error
                    # attempting to import it, otherwise we want it to bubble up.
                    curr = mod
                    curr_path = app
                    for part in item.split('.'):
                        if not module_has_submodule(curr, part):
                            break
                        curr_path += '.' + part
                        curr = import_module(curr_path)
                    else:
                        raise

    def ready(self):
        self.autodiscover()



# coding: utf-8
"""
This module contains base serializers to be used with django-rest-easy.

Crucial point of creating a good API is format consistency. If you've been lacking that so far, can't afford it anymore
or want to make your life easier, you can enforce a common message format and a common serialization format.
Enter the following SerializerCreator - it will make sure that everything serializers output will contain schema
and model fields. This affects both regular and model serializers.

Additional benefit of using such metaclass is serializer registration - we can easily obtain serializers based on
model (or None for non-model serializers) and schema from anywhere in the application. That's useful in several cases:
* model serialization
* remote data deserialization (no changes to (de)serialization logic required when we add a new schema)
* simpler views and viewsets

This doesn't disable any DRF's serializers functionality.
"""
from __future__ import unicode_literals

import six
from django.db import models
from rest_framework.serializers import (Serializer as OSerializer,
                                        ModelSerializer as OModelSerializer,
                                        SerializerMetaclass,
                                        Field)

from rest_easy.fields import StaticField
from rest_easy.registers import serializer_register
from rest_easy.patterns import RegisteredCreator

__all__ = ['ModelSerializer', 'Serializer', 'RegisterableSerializerMixin', 'SerializerCreator']


class SerializerCreator(RegisteredCreator, SerializerMetaclass):
    """
    This metaclass creates serializer classes to be used with django-rest-easy.

    We need to employ multiple inheritance here (if the behaviour ever needs to be overridden, you can just use both
    base classes to implement your own functionality) to preserve DRF's behaviour regarding
    serializer fields as well as registration and required fields checking from our own metaclass.

    Remember that all __new__ methods from base classes get called.
    """
    inherit_fields = False
    register = serializer_register
    required_fields = {
        'Meta': {
            'model': lambda value: value is None or issubclass(value, models.Model),
            'schema': lambda value: isinstance(value, six.string_types)
        }
    }

    @staticmethod
    def get_fields_from_base(base):
        """
        Alteration of original fields inheritance.

        It skips all serializer fields, since SerializerMetaclass deals with that already.
        :param base: a base class.
        :return: generator of (name, value) tuples of fields from base.
        """
        for item in dir(base):
            # Avoid copying serializer fields to class, since DRF's metaclass deals with that already.
            if not item.startswith('_') and not isinstance(item, Field):
                value = getattr(base, item)
                if not callable(value):
                    yield item, getattr(base, item)

    @staticmethod
    def get_name(name, bases, attrs):
        """
        Alteration of original get_name.

        This, instead of returing class's name, obtains correct serializer registration name from
        :class:`rest_easy.registers.SerializerRegister` and uses it as slug for registration purposes.
        :param name: class name.
        :param bases: class bases.
        :param attrs: class attributes.
        :return: registered serializer name.
        """
        model = attrs['Meta'].model
        return serializer_register.get_name(model, attrs['Meta'].schema)

    @classmethod
    def pre_register(mcs, name, bases, attrs):
        """
        Pre-register hook adding required fields

        This is the place to add required fields if they haven't been declared explicitly.
        We're adding model and schema fields here.
        :param name: class name.
        :param bases: class bases.
        :param attrs: class attributes.
        :return: tuple of altered name, bases, attrs.
        """
        if 'model' not in attrs:
            model = attrs['Meta'].model
            if model:
                model_name = '{}.{}'.format(model._meta.app_label, model._meta.object_name)  # pylint: disable=protected-access
            else:
                model_name = None
            attrs['model'] = StaticField(model_name)
        if 'schema' not in attrs:
            attrs['schema'] = StaticField(attrs['Meta'].schema)
        if  hasattr(attrs['Meta'], 'fields'):
            if not isinstance(attrs['Meta'].fields, six.string_types):
                attrs['Meta'].fields = list(attrs['Meta'].fields)
                if 'model' not in attrs['Meta'].fields:
                    attrs['Meta'].fields.append('model')
                if 'schema' not in attrs['Meta'].fields:
                    attrs['Meta'].fields.append('schema')
        return name, bases, attrs


class RegisterableSerializerMixin(object):  # pylint: disable=too-few-public-methods
    """
    A mixin to be used if you want to inherit functionality from non-standard DRF serializer.
    """
    __metaclass__ = SerializerCreator
    __abstract__ = True


class Serializer(OSerializer):  # pylint: disable=too-few-public-methods,abstract-method
    """
    Registered version of DRF's Serializer.
    """
    __metaclass__ = SerializerCreator
    __abstract__ = True


class ModelSerializer(OModelSerializer):  # pylint: disable=too-few-public-methods,abstract-method
    """
    Registered version of DRF's ModelSerializer.
    """
    __metaclass__ = SerializerCreator
    __abstract__ = True



# coding: utf-8
"""
Tests for django-rest-easy. So far not ported from proprietary code.
"""
from __future__ import unicode_literals

import unittest


class NullTest(unittest.TestCase):
    """
    Null test to check travis build.
    """
    def test_null(self):
        """
        As above.
        """
        self.assertTrue(True)  # pylint: disable=redundant-unittest-assert

if __name__ == '__main__':
    unittest.main()



"""
This class defines generic bases for a few design / architectural patterns
required byt django-rest-easy, namely singleton and register.
"""

from functools import wraps
from six import with_metaclass


class SingletonCreator(type):
    """
    This metaclass wraps __init__ method of created class with singleton_decorator.
    It's to make sure that it's impossible to mess up the instance for example by
    calling __init__ with getattr.
    """

    @staticmethod
    def singleton_decorator(function):
        """
        We embed given function into checking if the first (zeroth) parameter of its call
        shall be initialised.
        :param function: instantiating function (usually __init__).
        :returns: embedded function function.
        """

        @wraps(function)
        def wrapper(*args, **kwargs):
            """
            This inner function checks init property of given instance and depending on its
            value calls the function or not.
            """
            if args[0].sl_init:
                return function(*args, **kwargs)

        return wrapper

    def __new__(mcs, name, bases, attrs):
        """
        Wraps are awesome. Sometimes.
        """
        if not (len(bases) == 1 and object in bases):
            if '__init__' in attrs:
                attrs['__init__'] = mcs.singleton_decorator(attrs['__init__'])
        return super(SingletonCreator, mcs).__new__(mcs, name, bases, attrs)


class SingletonBase(object):  # pylint: disable=too-few-public-methods
    """
    This class implements the singleton pattern using a metaclass and
    overriding default __new__ magic method's behaviour. It works together with
    SingletonCreator metaclass to create a Singleton base class.
    sl_init property is reserved, you can't use it in inheriting classes.
    """

    _instance = None

    def __new__(cls, *_):
        """
        This magic method override makes sure that only one instance will be created.
        """
        if not isinstance(cls._instance, cls):
            cls._instance = super(SingletonBase, cls).__new__(cls)
            cls._instance.sl_init = True
        else:
            cls._instance.sl_init = False
        return cls._instance


class Singleton(with_metaclass(SingletonCreator, SingletonBase)):  # pylint: disable=too-few-public-methods
    """
    This is a Singleton you can inherit from.
    It reserves sl_init instance attribute to work properly.
    """
    pass


class BaseRegister(Singleton):
    """
    This class is a base register-type class. You should inherit from it to create particular registers.

    conflict_policy is a setting deciding what to do in case of name collision (registering another
    entity with the same name). It should be one of:
    * allow - replace old entry with new entry, return True,
    * deny - leave old entry, return False,
    * raise - raise ValueError.

    Default policy is raise.

    As this is a singleton, instantiating a particular children class in any place will yield the exact same data
    as the register instance used in RegisteredCreator().
    """
    conflict_policy = 'raise'

    def __init__(self):
        """
        We create an empty model dict.
        """
        self._entries = {}
        self.connect = lambda: None

    def register(self, name, ref):
        """
        Register an entry, shall we?
        :param name: entry name.
        :param ref: entry value (probably class).
        :returns: True if model was added just now, False if it was already in the register.
        """
        if not self.lookup(name) or self.conflict_policy == 'allow':
            self._entries[name] = ref
            return True
        elif self.conflict_policy == 'raise':
            raise ValueError('Entry named {} is already registered.'.format(name))
        return False

    def lookup(self, name):
        """
        I like to know if an entry is in the register, don't you?
        :param name: name to check.
        :returns: True if entry with given name is in the register, False otherwise.
        """
        return self._entries.get(name, None)

    def entries(self):
        """
        Return an iterator over all registered entries.
        """
        return self._entries.items()


class RegisteredCreator(type):
    """
    This metaclass integrates classes with a BaseRegister subclass.

    It skips processing base/abstract classes, which have __abstract__ property
    evaluating to True.
    """
    register = None
    required_fields = set()
    inherit_fields = False

    @staticmethod
    def get_name(name, bases, attrs):  # pylint: disable=unused-argument
        """
        Get name to be used for class registration.
        """
        return name

    @staticmethod
    def get_fields_from_base(base):
        """
        Obtains all fields from the base class.
        :param base: base class.
        :return: generator of (name, value) tuples.
        """
        for item in dir(base):
            if not item.startswith('_'):
                value = getattr(base, item)
                if not callable(value):
                    yield item, getattr(base, item)

    @classmethod
    def get_missing_fields(mcs, required_fields, fields):
        """
        Lists required fields that are missing.

        Supports two formats of input of required fields: either a simple set {'a', 'b'} or a dict with several
        options:
        {
            'nested': {
                'presence_check_only': None,
                'functional_check': lambda value: isinstance(value, Model)
            },
            'flat_presence_check': None,
            'flat_functional_check': lambda value: isinstance(value, Model)
        }
        Functional checks need to return true for field not to be marked as missing.
        Dict-format also supports both dict and attribute based accesses for fields (fields['a'] and fields.a).

        :param required_fields: set or dict of required fields.
        :param fields: dict or object of actual fields.
        :return: List of missing fields.
        """
        if isinstance(required_fields, set):
            return [field for field in required_fields if field not in fields or not field]

        missing = []
        for name, value in required_fields.iteritems():
            try:
                if not hasattr(fields, name) and name not in fields:
                    missing.append(name)
                    continue
            except TypeError:
                missing.append(name)
                continue
            if value:
                if hasattr(fields, name):
                    inner = getattr(fields, name)
                else:
                    inner = fields[name]
                if callable(value):
                    if not value(inner):
                        missing += [name]
                else:
                    missing += [name + '.' + item for item in mcs.get_missing_fields(value, inner)]
        return missing

    @classmethod
    def pre_register(mcs, name, bases, attrs):
        """
        Pre-register hook.
        :param name: class name.
        :param bases: class bases.
        :param attrs: class attributes.
        :return: Modified tuple (name, bases, attrs)
        """
        return name, bases, attrs

    @classmethod
    def post_register(mcs, cls, name, bases, attrs):
        """
        Post-register hook.
        :param cls: created class.
        :param name: class name.
        :param bases: class bases.
        :param attrs: class attributes.
        :return: None.
        """
        pass

    def __new__(mcs, name, bases, attrs):
        """
        This method creates and registers new class, if it's not already
        in the register.
        """
        # Do not register the base classes, which actual classes inherit.
        if mcs.inherit_fields:
            for base in bases:
                for field, value in mcs.get_fields_from_base(base):
                    if field not in attrs:
                        attrs[field] = value
        if not attrs.get('__abstract__', False):
            missing = mcs.get_missing_fields(mcs.required_fields, attrs)
            if missing:
                raise AttributeError(
                    'The following mandatory fields are missing from {} class definition: {}'.format(
                        name,
                        ', '.join(missing)
                    )
                )
            name, bases, attrs = mcs.pre_register(name, bases, attrs)
            slug = mcs.get_name(name, bases, attrs)
            cls = mcs.register.lookup(slug)
            if not cls:
                cls = super(RegisteredCreator, mcs).__new__(mcs, name, bases, attrs)
                mcs.register.register(slug, cls)
                mcs.post_register(cls, name, bases, attrs)
        else:
            cls = super(RegisteredCreator, mcs).__new__(mcs, name, bases, attrs)
        return cls



from setuptools import setup

setup(
    name='django-rest-easy',
    packages=['rest_easy'],
    version='0.1.1',
    python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*, !=3.3.*, <4',
    install_requires=[
        'django>=1.8.0',
        'djangorestframework>=3.0.0',
        'setuptools>=36.0.1',
        'six',
    ],
    description='django-rest-easy is an extension to DRF providing QOL improvements to serializers and views.',
    long_description='django-rest-easy enables:'
                     ' * versioning serializers by model and schema,'
                     ' * creating views and viewsets using model and schema,'
                     ' * serializer override for a particular DRF verb, like create or update,'
                     ' * scoping views\' querysets and viewsets by url kwargs or request object parameters.',
    author='SMARTPAGER SYSTEMS INC. / Krzysztof Bujniewicz',
    author_email='racech@gmail.com',
    url='https://github.com/TelmedIQ/django-rest-easy',
    keywords=['django', 'DRF', 'rest framework', 'serializers', 'viewsets'],
    license='MIT',
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
    ]
)



# coding: utf-8
"""
This module provides redefined DRF's generic views and viewsets leveraging serializer registration.

One of the main issues with creating traditional DRF APIs is a lot of bloat (and we're writing Python, not Java or C#,
to avoid bloat) that's completely unnecessary in a structured Django project. Therefore this module aims to provide
a better and simpler way to write simple API endpoints - without limiting the ability to create more complex views.
The particular means to that end are:
* :class:`rest_easy.views.ScopeQuerySet` and its subclasses (
  :class:`rest_easy.views.UrlKwargScopeQuerySet` and
  :class:`rest_easy.views.RequestAttrScopeQuerySet`) provide a simple way to scope views and viewsets
  by resource (ie. limiting results to single account, or /resource/<resource_pk>/inner_resource/<inner_resource_pk>/)
* generic views leveraging the above, as well as model-and-schema specification instead of queryset, serializer and
  helper methods - all generic views that were available in DRF as well as GenericAPIView are redefined to support
  this.
* Generic :class:`rest_easy.views.ModelViewSet` which allows for very simple definition of resource
  endpoint.

To make the new views work, all that's required is a serializer:

```python
from users.models import User
from accounts.models import Account
from rest_easy.serializers import ModelSerializer
class UserSerializer(ModelSerializer):
    class Meta:
        model = User
        fields = '__all__'
        schema = 'default'

class UserViewSet(ModelViewSet):
    model = User
    scope = UrlKwargScopeQuerySet(Account)
```

and in urls.py:

```python
from django.conf.urls import url, include
from rest_framework.routers import DefaultRouter
router = DefaultRouter()
router.register(r'accounts/(?P<account_pk>[0-9]+)/users', UserViewSet)
urlpatterns = [url(r'^', include(router.urls))]
```

The above will provide the users scoped by account primary key as resources: with list, retrieve, create, update and
partial update methods, as well as standard HEAD and OPTIONS autogenerated responses.

You can easily add custom paths to viewsets when needed - it's described in DRF documentation.
"""
from __future__ import unicode_literals

from django.conf import settings
from django.db.models import QuerySet, Model
from django.http import Http404
from django.shortcuts import get_object_or_404
from rest_framework.viewsets import ViewSetMixin
from rest_framework import generics, mixins
from six import with_metaclass

from rest_easy.exceptions import RestEasyException
from rest_easy.registers import serializer_register


class ScopeQuerySet(object):
    """
    This class provides a scope-by-parent-element functionality to views and their querysets.

    It works by selecting a proper parent model instance and filtering view's queryset with it automatically.
    """

    def __init__(self, qs_or_obj, parent_field='id', related_field=None, raise_404=False, parent=None,
                 allow_none=False):
        """
        Sets instance properties, infers sane defaults and ensures qs_or_obj is correct.

        :param qs_or_obj: This can be a queryset or a Django model or explicit None (for particular subclasses)
        :param parent_field: the field to filter by in the parent queryset (qs_or_obj), by default 'id'.
        :param related_field: the field to filter by in the view queryset, by default model_name.
        :param raise_404: whether 404 should be raised if parent object cannot be found.
        :param parent: if this object's queryset should be filtered by another parameter, parent attribute should be
         an instance of ScopeQuerySet. This allows for ScopeQuerySetChaining (ie. for messages we might have
         UrlKwargScopeQuerySet(User, parent=UrlKwargScopeQuerySet(Account))  for scoping by user and limiting users
         to an account.
        :param allow_none: if filtering view queryset by object=None should be allowed. If it's false, resulting
         queryset is guaranteed to be empty if parent object can't be found and 404 is not raised.
        """
        if isinstance(qs_or_obj, QuerySet):
            self.queryset = qs_or_obj
        elif issubclass(qs_or_obj, Model):
            self.queryset = qs_or_obj.objects.all()
        elif qs_or_obj is None:
            self.queryset = None
        else:
            raise RestEasyException('Queryset parameter must be an instance of QuerySet or a Model subclass.')

        if related_field is None:
            try:
                related_field = '{}'.format(self.queryset.model._meta.model_name)  # pylint: disable=protected-access
            except AttributeError:
                raise RestEasyException('Either related_field or qs_or_obj must be given.')
        self.parent_field = parent_field
        self.related_field = related_field
        self.raise_404 = raise_404
        self.parent = [parent] if isinstance(parent, ScopeQuerySet) else parent
        self.allow_none = allow_none

    def get_value(self, view):
        """
        Get value used to filter qs_or_objs's field specified for filtering (parent_field in init).
        :param view: DRF view instance - as it provides access to both request and kwargs.
        :return: value to filter by.
        """
        raise NotImplementedError('You need to use ScopeQueryset subclass with get_value implemented.')

    def get_queryset(self, view):
        """
        Obtains parent queryset (init's qs_or_obj) along with any chaining (init's parent) required.
        :param view: DRF view instance.
        :return: queryset instance.
        """
        queryset = self.queryset
        if self.parent:
            queryset = self.parent.child_queryset(queryset, view)
        return queryset

    def get_object(self, view):
        """
        Obtains parent object by which view queryset should be filtered.
        :param view: DRF view instance.
        :return: object (instance of init's qs_or_obj model except shadowed by subclass).
        """
        queryset = self.get_queryset(view)
        queryset = queryset.filter(**{self.parent_field: self.get_value(view)})
        try:
            obj = get_object_or_404(queryset)
        except Http404:
            if self.raise_404:
                raise
            obj = None
        return obj

    def child_queryset(self, queryset, view):
        """
        Performs filtering of the view queryset.
        :param queryset: view queryset instance.
        :param view: view object.
        :return: filtered queryset.
        """
        obj = self.get_object(view)
        if obj is None and not self.allow_none:
            return queryset.none()
        return queryset.filter(**{self.related_field: obj})


class UrlKwargScopeQuerySet(ScopeQuerySet):
    """
    ScopeQuerySet that obtains parent object from url kwargs.
    """

    def __init__(self, *args, **kwargs):
        """
        Adds url_kwarg to :class:`rest_easy.views.ScopeQuerySet` init parameters.

        :param args: same as :class:`rest_easy.views.ScopeQuerySet`.
        :param url_kwarg: name of url field to be obtained from view's kwargs. By default it will be inferred as
         model_name_pk.
        :param kwargs: same as :class:`rest_easy.views.ScopeQuerySet`.
        """
        self.url_kwarg = kwargs.pop('url_kwarg', None)
        super(UrlKwargScopeQuerySet, self).__init__(*args, **kwargs)
        if not self.url_kwarg:
            try:
                self.url_kwarg = '{}_pk'.format(self.queryset.model._meta.model_name)  # pylint: disable=protected-access
            except AttributeError:
                raise RestEasyException('Either related_field or qs_or_obj must be given.')

    def get_value(self, view):
        """
        Obtains value from url kwargs.
        :param view: DRF view instance.
        :return: Value determining parent object.
        """
        return view.kwargs.get(self.url_kwarg)


class RequestAttrScopeQuerySet(ScopeQuerySet):
    """
    ScopeQuerySet that obtains parent object from view's request property.

    It can work two-fold:
    * the request's property contains full object: in this case no filtering of parent's queryset is required. When
     using such approach, is_object must be set to True, and qs_or_obj can be None. Chaining will be disabled since it
     is inherent to filtering process.
    * the request's property contains object's id, uuid, or other unique property. In that case is_object needs to be
     explicitly set to False, and qs_or_obj needs to be a Django model or queryset. Chaining will be performed as
     usually.
    """

    def __init__(self, *args, **kwargs):
        """
        Adds is_object and request_attr  to :class:`rest_easy.views.ScopeQuerySet` init parameters.

        :param args: same as :class:`rest_easy.views.ScopeQuerySet`.
        :param request_attr: name of property to be obtained from view.request.
        :param is_object: if request's property will be an object or a value to filter by. True by default.
        :param kwargs: same as :class:`rest_easy.views.ScopeQuerySet`.
        """
        self.request_attr = kwargs.pop('request_attr', None)
        if self.request_attr is None:
            raise RestEasyException('request_attr must be set explicitly on an {} init.'.format(
                self.__class__.__name__))
        self.is_object = kwargs.pop('is_object', True)
        super(RequestAttrScopeQuerySet, self).__init__(*args, **kwargs)

    def get_value(self, view):
        """
        Obtains value from url kwargs.
        :param view: DRF view instance.
        :return: Value determining parent object.
        """
        return getattr(view.request, self.request_attr, None)

    def get_object(self, view):
        """
        Extends standard get_object's behaviour with handling values that are already objects.
        :param view: DRF view instance.
        :return: object to filter view's queryset by.
        """
        if self.is_object:
            return self.get_value(view)
        return super(RequestAttrScopeQuerySet, self).get_object(view)


class ScopedViewMixin(object):
    """
    This class provides a get_queryset method that works with ScopeQuerySet.

    Queryset obtained from superclass is filtered by view.scope's (if it exists) child_queryset() method.
    """

    def get_queryset(self):
        """
        Calls scope's child_queryset methods on queryset as obtained from superclass.
        :return: queryset.
        """
        queryset = super(ScopedViewMixin, self).get_queryset()
        if hasattr(self, 'scope'):
            if isinstance(self.scope, ScopeQuerySet):
                queryset = self.scope.child_queryset(queryset, self)
            else:
                for scope in self.scope:
                    queryset = scope.child_queryset(queryset, self)
        return queryset


class ViewEasyMetaclass(type):
    """
    This metaclass sets default queryset on a model-and-schema based views and fills in concrete views with bases.

    It's required for compatibility with some of DRF's elements, like routers.
    """
    resolved_bases = None

    @classmethod
    def get_additional_bases(mcs):
        """
        Looks for additional view bases in settings.REST_EASY_VIEW_BASES
        :return:
        """
        if mcs.resolved_bases is None:
            mcs.resolved_bases = []
            from importlib import import_module
            for base in getattr(settings, 'REST_EASY_VIEW_BASES'):
                mod, cls = base.rsplit('.', 1)
                mcs.resolved_bases.append(getattr(import_module(mod), cls))

        return mcs.resolved_bases

    def __new__(mcs, name, bases, attrs):
        """
        Create the class.
        """
        if ('queryset' not in attrs or attrs['queryset'] is None) and 'model' in attrs:
            attrs['queryset'] = attrs['model'].objects.all()
        if (not attrs.get('__abstract__', False)) and 'parent' in attrs and isinstance(attrs['parent'], ScopeQuerySet):
            attrs['parent'] = [attrs['parent']]
        bases = tuple(mcs.get_additional_bases() + list(bases))
        return super(ViewEasyMetaclass, mcs).__new__(mcs, name, bases, attrs)


class ChainingCreateUpdateMixin(object):
    """
    Chain-enabled versions of perform_create and perform_update.
    """
    def perform_create(self, serializer, **kwargs):  # pylint: disable=no-self-use
        """
        Extend default implementation with kwarg chaining.
        """
        return serializer.save(**kwargs)

    def perform_update(self, serializer, **kwargs):  # pylint: disable=no-self-use
        """
        Extend default implementation with kwarg chaining.
        """
        return serializer.save(**kwargs)


class GenericAPIViewBase(ScopedViewMixin, generics.GenericAPIView):
    """
    Provides a base for all generic views and viewsets leveraging registered serializers and ScopeQuerySets.

    Adds additional DRF-verb-wise override for obtaining serializer class: serializer_schema_for_verb property.
    It should be a dictionary of DRF verbs and serializer schemas (they work in conjunction with model property).
    The priority for obtaining serializer class is:
    * get_serializer_class override
    * serializer_class property
    * model + serializer_schema_for_verb[verb] lookup in :class:`rest_easy.registers.SerializerRegister`
    * model + schema lookup in :class:`rest_easy.registers.SerializerRegister`
    """
    serializer_schema_for_verb = {}

    def get_drf_verb(self):
        """
        Obtain the DRF verb used for a request.
        """
        method = self.request.method.lower()
        if method == 'get':
            if self.lookup_url_kwarg in self.kwargs:
                return 'retrieve'
            return 'list'
        mapping = {
            'post': 'create',
            'put': 'update',
            'patch': 'partial_update',
            'delete': 'destroy'
        }
        return mapping[method]

    def get_serializer_name(self, verb=None):
        """
        Obtains registered serializer name for this view.

        Leverages :class:`rest_easy.registers.SerializerRegister`. Works when either of or both model
        and schema properties are available on this view.

        :return: registered serializer key.
        """
        model = getattr(self, 'model', None)
        schema = None
        if not model and not hasattr(self, 'schema') and (verb and verb not in self.serializer_schema_for_verb):
            raise RestEasyException('Either model or schema fields need to be set on a model-based GenericAPIView.')
        if verb:
            schema = self.serializer_schema_for_verb.get(verb, None)
        if schema is None:
            schema = getattr(self, 'schema', 'default')
        return serializer_register.get_name(model, schema)

    def get_serializer_class(self):
        """
        Gets serializer appropriate for this view.

        Leverages :class:`rest_easy.registers.SerializerRegister`. Works when either of or both model
        and schema properties are available on this view.

        :return: serializer class.
        """

        if hasattr(self, 'serializer_class') and self.serializer_class:
            return self.serializer_class

        serializer = serializer_register.lookup(self.get_serializer_name(verb=self.get_drf_verb()))
        if serializer:
            return serializer

        raise RestEasyException('Serializer for model {} and schema {} cannot be found.'.format(
            getattr(self, 'model', '[no model]'),
            getattr(self, 'schema', '[no schema]')
        ))


class GenericAPIView(with_metaclass(ViewEasyMetaclass, GenericAPIViewBase)):
    """
    Base view with compat metaclass.
    """
    __abstract__ = True


class CreateAPIView(ChainingCreateUpdateMixin,
                    mixins.CreateModelMixin,
                    GenericAPIView):
    """
    Concrete view for creating a model instance.
    """

    def post(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.create(request, *args, **kwargs)


class ListAPIView(mixins.ListModelMixin,
                  GenericAPIView):
    """
    Concrete view for listing a queryset.
    """

    def get(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.list(request, *args, **kwargs)


class RetrieveAPIView(mixins.RetrieveModelMixin,
                      GenericAPIView):
    """
    Concrete view for retrieving a model instance.
    """

    def get(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.retrieve(request, *args, **kwargs)


class DestroyAPIView(mixins.DestroyModelMixin,
                     GenericAPIView):
    """
    Concrete view for deleting a model instance.
    """

    def delete(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.destroy(request, *args, **kwargs)


class UpdateAPIView(ChainingCreateUpdateMixin,
                    mixins.UpdateModelMixin,
                    GenericAPIView):
    """
    Concrete view for updating a model instance.
    """

    def put(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.partial_update(request, *args, **kwargs)


class ListCreateAPIView(ChainingCreateUpdateMixin,
                        mixins.ListModelMixin,
                        mixins.CreateModelMixin,
                        GenericAPIView):
    """
    Concrete view for listing a queryset or creating a model instance.
    """

    def get(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.list(request, *args, **kwargs)

    def post(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.create(request, *args, **kwargs)


class RetrieveUpdateAPIView(ChainingCreateUpdateMixin,
                            mixins.RetrieveModelMixin,
                            mixins.UpdateModelMixin,
                            GenericAPIView):
    """
    Concrete view for retrieving, updating a model instance.
    """

    def get(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.partial_update(request, *args, **kwargs)


class RetrieveDestroyAPIView(mixins.RetrieveModelMixin,
                             mixins.DestroyModelMixin,
                             GenericAPIView):
    """
    Concrete view for retrieving or deleting a model instance.
    """

    def get(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.retrieve(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.destroy(request, *args, **kwargs)


class RetrieveUpdateDestroyAPIView(ChainingCreateUpdateMixin,
                                   mixins.RetrieveModelMixin,
                                   mixins.UpdateModelMixin,
                                   mixins.DestroyModelMixin,
                                   GenericAPIView):
    """
    Concrete view for retrieving, updating or deleting a model instance.
    """

    def get(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.retrieve(request, *args, **kwargs)

    def put(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.update(request, *args, **kwargs)

    def patch(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.partial_update(request, *args, **kwargs)

    def delete(self, request, *args, **kwargs):
        """
        Shortcut method.
        """
        return self.destroy(request, *args, **kwargs)


class GenericViewSet(ViewSetMixin, GenericAPIView):
    """
    The GenericViewSet class does not provide any actions by default,
    but does include the base set of generic view behavior, such as
    the `get_object` and `get_queryset` methods.
    """
    pass


class ReadOnlyModelViewSet(mixins.RetrieveModelMixin,
                           mixins.ListModelMixin,
                           GenericViewSet):
    """
    A viewset that provides default `list()` and `retrieve()` actions.
    """
    pass


class ModelViewSet(ChainingCreateUpdateMixin,
                   mixins.CreateModelMixin,
                   mixins.RetrieveModelMixin,
                   mixins.UpdateModelMixin,
                   mixins.DestroyModelMixin,
                   mixins.ListModelMixin,
                   GenericViewSet):
    """
    A viewset that provides default `create()`, `retrieve()`, `update()`,
    `partial_update()`, `destroy()` and `list()` actions.
    """
    pass
